"""aPRAW constants."""
from .endpoints import API_PATH, BASE_URL  # noqa: F401

__version__ = "0.6.9"
__tag__ = "alpha"

USER_AGENT_FORMAT = "{} aPRAW v" + __version__ + "-" + __tag__ + "by Dan6erbond"
